from __future__ import annotations
import asyncio
import itertools
from enum import Enum
from typing import Optional

from aiobotocore.session import get_session  # type: ignore

from near_lake_framework import near_primitives
from near_lake_framework import s3_fetchers


from dataclasses import dataclass


class Network(Enum):
    MAINNET = "mainnet"
    TESTNET = "testnet"

    @staticmethod
    def from_string(value: str):
        try:
            return Network(value.lower())
        except ValueError as err:
            valid_values = [v.value for v in Network]
            raise ValueError(
                f"Unknown network: {value}. Valid values are: {valid_values}"
            ) from err

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return self.value


@dataclass
class LakeConfig:
    s3_bucket_name: str
    s3_region_name: str
    aws_access_key_id: str
    aws_secret_key: str
    start_block_height: near_primitives.BlockHeight
    blocks_preload_pool_size: int = 200

    def __init__(
        self,
        network: Network,
        aws_access_key_id: str,
        aws_secret_key: str,
        start_block_height: near_primitives.BlockHeight,
        preload_pool_size: int = 200,
    ):
        # These are entirely determined by Network.
        self.s3_bucket_name = f"near-lake-data-{network.value}"
        self.s3_region_name = "eu-central-1"

        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_key = aws_secret_key
        self.start_block_height = start_block_height
        self.preload_pool_size = preload_pool_size


async def start(config: LakeConfig, streamer_messages_queue: asyncio.Queue):
    session = get_session()
    async with session.create_client(
        "s3",
        region_name=config.s3_region_name,
        aws_secret_access_key=config.aws_secret_key,
        aws_access_key_id=config.aws_access_key_id,
    ) as s3_client:
        start_from_block_height: near_primitives.BlockHeight = config.start_block_height
        last_processed_block_hash: Optional[str] = None

        while True:
            block_heights_prefixes = await s3_fetchers.list_blocks(
                s3_client,
                config.s3_bucket_name,
                start_from_block_height,
                config.blocks_preload_pool_size * 2,
            )

            if not block_heights_prefixes:
                print("No new blocks on S3, retry in 2s...")

                await asyncio.sleep(2)
                continue

            print("Received {} blocks from S3".format(len(block_heights_prefixes)))

            pending_block_heights = iter(block_heights_prefixes)
            streamer_messages_futures = []
            for block_height in itertools.islice(
                pending_block_heights, streamer_messages_queue.maxsize
            ):
                streamer_messages_futures.append(
                    asyncio.create_task(
                        s3_fetchers.fetch_streamer_message(
                            s3_client, config.s3_bucket_name, block_height
                        )
                    )
                )

            while streamer_messages_futures:
                streamer_message = await streamer_messages_futures.pop(0)

                if (
                    last_processed_block_hash
                    and last_processed_block_hash
                    != streamer_message.block.header.prev_hash
                ):
                    print(
                        "`prev_hash` does not match, refetching the data from S3 in 200ms",
                        last_processed_block_hash,
                        streamer_message.block.header.prev_hash,
                    )

                    await asyncio.sleep(0.2)
                    break

                last_processed_block_hash = streamer_message.block.header.hash
                start_from_block_height = streamer_message.block.header.height + 1

                try:
                    pending_block_height = next(pending_block_heights)
                except StopIteration:
                    pass
                else:
                    streamer_messages_futures.append(
                        asyncio.create_task(
                            s3_fetchers.fetch_streamer_message(
                                s3_client,
                                config.s3_bucket_name,
                                pending_block_height,
                            )
                        )
                    )

                await streamer_messages_queue.put(streamer_message)


def streamer(config: LakeConfig):
    streamer_messages_queue: asyncio.Queue = asyncio.Queue(
        maxsize=config.blocks_preload_pool_size
    )
    stream_handle = asyncio.create_task(start(config, streamer_messages_queue))
    return stream_handle, streamer_messages_queue



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="near-lake-framework",
    version="0.0.8",
    author="Ivan Frolov",
    author_email="frolvanya@gmail.com",
    description="Python Library to connect to the NEAR Lake S3 and stream the data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/frolvanya/near-lake-framework-py",
    project_urls={
        "Bug Tracker": "https://github.com/frolvanya/near-lake-framework-py/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.9",
    install_requires=[
        "asyncio>=3.4.3",
        "dataclasses>=0.6",
        "dataclasses-json>=0.5.7",
        "aiobotocore>=2.3.0",
    ],
    package_data={"near_lake_framework": ["py.typed"]},
)



from typing import Any, List, Optional

from dataclasses import dataclass, field
from dataclasses_json import DataClassJsonMixin, config, mm


AccountId = str
CryptoHash = str
BlockHeight = int


class BlockHeightField(mm.fields.Integer):
    """Block Height is unsigned 64-bit integer, so it needs to be serialized as a string and get deserialized
    to an integer type in Python.
    """

    def __init__(self, *args, **kwargs):
        return super().__init__(*args, **kwargs, as_string=True)


@dataclass
class BlockHeader(DataClassJsonMixin):
    epoch_id: CryptoHash
    next_epoch_id: CryptoHash
    hash: CryptoHash
    prev_hash: CryptoHash
    prev_state_root: CryptoHash
    chunk_receipts_root: CryptoHash
    chunk_headers_root: CryptoHash
    chunk_tx_root: CryptoHash
    outcome_root: CryptoHash
    chunks_included: int
    challenges_root: CryptoHash
    timestamp: int
    timestamp_nanosec: int = field(
        metadata=config(mm_field=mm.fields.Integer(as_string=True))
    )
    random_value: CryptoHash
    validator_proposals: List[Any]
    chunk_mask: List[bool]
    gas_price: int = field(metadata=config(mm_field=mm.fields.Integer(as_string=True)))
    block_ordinal: Optional[int]
    rent_paid: int = field(metadata=config(mm_field=mm.fields.Integer(as_string=True)))
    validator_reward: int = field(
        metadata=config(mm_field=mm.fields.Integer(as_string=True))
    )
    total_supply: int = field(
        metadata=config(mm_field=mm.fields.Integer(as_string=True))
    )
    challenges_result: List[Any]
    last_final_block: CryptoHash
    last_ds_final_block: CryptoHash
    next_bp_hash: CryptoHash
    block_merkle_root: CryptoHash
    epoch_sync_data_hash: Optional[CryptoHash]
    approvals: List[Optional[str]]
    signature: str
    latest_protocol_version: int
    height: BlockHeight = field(metadata=config(mm_field=BlockHeightField()))
    prev_height: Optional[BlockHeight] = field(
        default=None, metadata=config(mm_field=BlockHeightField())
    )


@dataclass
class ChunkHeader(DataClassJsonMixin):
    chunk_hash: CryptoHash
    prev_block_hash: CryptoHash
    outcome_root: CryptoHash
    prev_state_root: CryptoHash
    encoded_merkle_root: CryptoHash
    encoded_length: int
    height_created: int
    height_included: int
    shard_id: int
    gas_used: int
    gas_limit: int
    rent_paid: int = field(metadata=config(mm_field=mm.fields.Integer(as_string=True)))
    validator_reward: int = field(
        metadata=config(mm_field=mm.fields.Integer(as_string=True))
    )
    balance_burnt: int = field(
        metadata=config(mm_field=mm.fields.Integer(as_string=True))
    )
    outgoing_receipts_root: CryptoHash
    tx_root: CryptoHash
    validator_proposals: List[Any]
    signature: str


@dataclass
class Block(DataClassJsonMixin):
    author: AccountId
    header: BlockHeader
    chunks: List[ChunkHeader]


@dataclass
class Receipt(DataClassJsonMixin):
    predecessor_id: AccountId
    receiver_id: AccountId
    receipt_id: CryptoHash
    receipt: Any


@dataclass
class CostGasUsed(DataClassJsonMixin):
    cost_category: str
    cost: str
    gas_used: int


@dataclass
class ExecutionMetadata(DataClassJsonMixin):
    version: int
    gas_profile: Optional[List[CostGasUsed]]


@dataclass
class ExecutionOutcome(DataClassJsonMixin):
    logs: List[str]
    receipt_ids: List[CryptoHash]
    gas_burnt: int
    tokens_burnt: int = field(
        metadata=config(mm_field=mm.fields.Integer(as_string=True))
    )
    executor_id: AccountId
    status: Any
    metadata: ExecutionMetadata


@dataclass
class ExecutionOutcomeWithId(DataClassJsonMixin):
    proof: List[Any]
    block_hash: CryptoHash
    id: CryptoHash
    outcome: ExecutionOutcome


@dataclass
class IndexerExecutionOutcomeWithReceipt(DataClassJsonMixin):
    execution_outcome: ExecutionOutcomeWithId
    receipt: Receipt


@dataclass
class SignedTransaction(DataClassJsonMixin):
    signer_id: AccountId
    public_key: str
    nonce: int
    receiver_id: AccountId
    actions: List[Any]
    signature: str
    hash: CryptoHash


@dataclass
class IndexerExecutionOutcomeWithOptionalReceipt:
    execution_outcome: ExecutionOutcomeWithId
    receipt: Optional[Receipt]


@dataclass
class IndexerTransactionWithOutcome(DataClassJsonMixin):
    transaction: SignedTransaction
    outcome: IndexerExecutionOutcomeWithOptionalReceipt


@dataclass
class IndexerChunk(DataClassJsonMixin):
    author: AccountId
    header: ChunkHeader
    transactions: List[IndexerTransactionWithOutcome]
    receipts: List[Receipt]


@dataclass
class IndexerShard(DataClassJsonMixin):
    shard_id: int
    chunk: Optional[IndexerChunk]
    receipt_execution_outcomes: List[IndexerExecutionOutcomeWithReceipt]
    state_changes: List[Any]


@dataclass
class StreamerMessage(DataClassJsonMixin):
    block: Block
    shards: List[IndexerShard]
